﻿using System;
using System.Collections.Generic;

using System.Net;

using RestSharp;
using Newtonsoft.Json;
using System.IO;

namespace SignorityAPI
{

    public class Signority
    {
 
        public String baseUrl = "http://test.signority.com/api/";
        
        public Dictionary<string, string> AdditionalParameters { get; set; } = new Dictionary<string, string>();

        private RestClient client;
        private RestSharp.Deserializers.JsonDeserializer deserializer;
        public List<Warning> Warnings { get; private set; }
       

        public Signority()
        {
            client = new RestClient();
            client.BaseUrl = new Uri(baseUrl);
            client.UserAgent = "Signority-CSharp-API-V1.0";

            deserializer = new RestSharp.Deserializers.JsonDeserializer();
            Warnings = new List<Warning>();

        }

        public Signority(string username, string password) : this()
        {
        //    this.apikey = apikey;
            //client.Authenticator = new RestSharp.Authenticators.HttpBasicAuthenticator(apiKey, "");
            client.Authenticator = new RestSharp.Authenticators.HttpBasicAuthenticator(username, password);
        }


        private void HandleErrors(IRestResponse response)
        {
            // If there was an exception getting the response
            if (response.ErrorException != null && response.ContentType != "text/html")
            {
                const string message = "Error info from response.";
                throw new ApplicationException(message, response.ErrorException);
            }

            if (response.ContentType == "application/json")
            {
                // Check for an error
                deserializer.RootElement = "error";
                var error = deserializer.Deserialize<Error>(response);
                if (error.ErrorName != null)
                {
                    switch (error.ErrorName)
                    {
                        case "bad_request":
                            throw new BadRequestException(error.ErrorMsg, error.ErrorName);
                        case "unauthorized":
                            throw new UnauthorizedException(error.ErrorMsg, error.ErrorName);
                        case "payment_required":
                            throw new PaymentRequiredException(error.ErrorMsg, error.ErrorName);
                        case "forbidden":
                            throw new ForbiddenException(error.ErrorMsg, error.ErrorName);
                        case "not_found":
                            throw new NotFoundException(error.ErrorMsg, error.ErrorName);
                        case "conflict":
                            throw new ConflictException(error.ErrorMsg, error.ErrorName);
                        case "team_invite_failed":
                            throw new ForbiddenException(error.ErrorMsg, error.ErrorName);
                        case "invalid_recipient":
                            throw new BadRequestException(error.ErrorMsg, error.ErrorName);
                        case "signature_request_cancel_failed":
                            throw new BadRequestException(error.ErrorMsg, error.ErrorName);
                        case "maintenance":
                            throw new ServiceUnavailableException(error.ErrorMsg, error.ErrorName);
                        case "deleted":
                            throw new GoneException(error.ErrorMsg, error.ErrorName);
                        case "unknown":
                            throw new UnknownException(error.ErrorMsg, error.ErrorName);
                        case "method_not_supported":
                            throw new MethodNotAllowedException(error.ErrorMsg, error.ErrorName);
                        case "signature_request_invalid":
                            throw new ErrorException(error.ErrorMsg, error.ErrorName);
                        case "template_error":
                            throw new ErrorException(error.ErrorMsg, error.ErrorName);
                        case "invalid_reminder":
                            throw new BadRequestException(error.ErrorMsg, error.ErrorName);
                        case "exceeded_rate":
                            throw new ExceededRateLimitException(error.ErrorMsg, error.ErrorName);
                        default:
                            throw new ErrorException(error.ErrorMsg, error.ErrorName);
                    }
                }

                // Look for warnings
                deserializer.RootElement = "warnings";
                var warnings = deserializer.Deserialize<List<Warning>>(response);
                if (warnings[0].WarningName != null)
                {
                    Warnings.AddRange(warnings);
                }
            }

            // Throw exceptions appropriate for the HTTP status code
            switch (response.StatusCode)
            {
                case HttpStatusCode.BadRequest:
                    throw new BadRequestException();
                case HttpStatusCode.Unauthorized:
                    throw new UnauthorizedException();
                case HttpStatusCode.PaymentRequired:
                    throw new PaymentRequiredException();
                case HttpStatusCode.Forbidden:
                    throw new ForbiddenException();
                case HttpStatusCode.NotFound:
                    throw new NotFoundException();
                case HttpStatusCode.MethodNotAllowed:
                    throw new MethodNotAllowedException();
                case HttpStatusCode.Conflict:
                    throw new ConflictException();
                case HttpStatusCode.Gone:
                    throw new GoneException();
                case HttpStatusCode.InternalServerError:
                    throw new UnknownException();
            }

            // Throw an exception for any non-2xx status code we didn't cover above
            var statusCode = (int)response.StatusCode;
            if ((statusCode < 200) || (statusCode >= 300))
            {
                throw new ApplicationException("Received status " + statusCode + " from server. Full response:\n" + response.Content);
            }
        }

        private void RequireAuthentication()
        {
            if (client.Authenticator == null)
            {
                throw new UnauthorizedAccessException("This method requires authentication");
            }
        }

        private void InjectAdditionalParameters(RestRequest request)
        {
            foreach (KeyValuePair<string, string> entry in this.AdditionalParameters)
            {
                request.AddParameter(entry.Key, entry.Value);
            }
        }


        private T Execute<T>(RestRequest request, string ver) where T : new()
        {

            client.BaseUrl = new Uri(baseUrl+ ver + "/");

            InjectAdditionalParameters(request);
            var response = client.Execute<T>(request);
            HandleErrors(response);
            return response.Data;
        }
               
        public Client GetAccount()
        {
            RequireAuthentication();

            var request = new RestRequest("user");
            request.RootElement = "user";
            return Execute<Client>(request, "v1");
        }

        public JsonResultDocumentList<Documents> GetDocumentList(int? page = null, int? pageSize = null)
        {
            RequireAuthentication();

            var request = new RestRequest("documents/search2"); 

            if (page == null)       page = 1;
            if (pageSize == null)   pageSize = 100;
            
            request.AddParameter("pageIndex", page);
            request.AddParameter("pageSize", pageSize);
            request.AddParameter("type", "Document");
            request.AddParameter("status", "allStatus");
            request.AddParameter("orderBy", "LastModifiedTime");
            request.AddParameter("orderSeq", "DESC");

            return Execute<JsonResultDocumentList<Documents>>(request, "v2");
        }

        public JsonResultDocument GetOneDocument(int docId)
        {
            RequireAuthentication();

            client.BaseUrl = new Uri(baseUrl);

            var request = new RestRequest("documents/" + docId);

            return Execute<JsonResultDocument>(request, "v1");
        }

        public JsonResultDocument sendDocument(Documents doc1, string[] files)
        {
            RequireAuthentication();

            var request = new RestRequest("documents", Method.POST);

            request.AddHeader("Accept", "application/json");
            request.Parameters.Clear();
            string json = request.JsonSerializer.Serialize(doc1);
            request.AddParameter("document", json);

            foreach (string f in files)
            {
                request.AddFile("files", f);
            }
           
            return Execute<JsonResultDocument>(request, "v1");
        }

        public JsonResultDocument sendDocument(DocumentAnchorTag doc1, string[] files)
        {
            RequireAuthentication();

            var request = new RestRequest("documents", Method.POST);

            request.AddHeader("Accept", "application/json");
            request.Parameters.Clear();
            string json = request.JsonSerializer.Serialize(doc1);
            request.AddParameter("document", json);

            foreach (string f in files)
            {
                request.AddFile("files", f);
            }

            return Execute<JsonResultDocument>(request, "v1");
        }

        public JsonResultDocument sendDocumentWithoutSignerTag(DocumentWithoutSignerTag doc1, string[] files)
        {
            RequireAuthentication();

            var request = new RestRequest("documents", Method.POST);

            request.AddHeader("Accept", "application/json");
            request.Parameters.Clear();
            string json = request.JsonSerializer.Serialize(doc1);
            request.AddParameter("document", json);

            foreach (string f in files)
            {
                request.AddFile("files", f);
            }

            return Execute<JsonResultDocument>(request, "v1");
        }

        public JsonResultDocument updateDocument(Documents doc1)
        {
            RequireAuthentication();

            var request = new RestRequest("documents/" + doc1.id, Method.PUT);
            request.Method = Method.PUT;

            request.AddHeader("Accept", "application/json");
            request.Parameters.Clear();
            string json = request.JsonSerializer.Serialize(doc1);
            request.AddParameter("document", json);

            return Execute<JsonResultDocument>(request, "v1");
        }

        public JsonResultDocument uploadFiles(int doc_id, string[] files)
        {
            RequireAuthentication();

            var request = new RestRequest("documents/" + doc_id + "/files.html", Method.POST);

            request.AddHeader("Accept", "text/html");
            request.AddHeader("Content-Type", "multipart/form-data");
            request.RequestFormat = DataFormat.Json;
            request.Parameters.Clear();
            

            request.AddParameter("addTagsOfPDF", false);
            request.AddParameter("addAnchorTags", false);
            request.AddParameter("highDisplayResolution", false);

            foreach (string f in files)
            {
                request.AddFile("files", f);
            }


            return Execute<JsonResultDocument>(request, "v1");
        }

        public JsonResultTags addTags(int doc_id, List<Tag> tags)
        {
            RequireAuthentication();

            var request = new RestRequest("documents/" + doc_id + "/tags", Method.PUT);
            request.Method = Method.PUT;

            request.AddHeader("Accept", "application/json");
            request.Parameters.Clear();
            string json = request.JsonSerializer.Serialize(tags);
            request.AddParameter("document", json);

            return Execute<JsonResultTags>(request, "v1");
        }

    }

}

